# SPDX-License-Identifier: AGPL-3.0-or-later
"""Functional, browser based tests for ejabberd app."""

import pytest

from plinth.tests import functional

pytestmark = [pytest.mark.apps, pytest.mark.ejabberd]

# TODO Check domain name displayed in description


class TestEjabberdApp(functional.BaseAppTests):
    app_name = 'ejabberd'
    has_service = True
    has_web = False

    @pytest.fixture(scope='module', autouse=True)
    def fixture_background(self, session_browser):
        """Login, install, and enable the app."""
        functional.login(session_browser)
        functional.install(session_browser, 'jsxc')
        functional.install(session_browser, self.app_name)
        functional.app_enable(session_browser, 'jsxc')
        functional.app_enable(session_browser, self.app_name)
        yield
        functional.login(session_browser)
        functional.app_disable(session_browser, self.app_name)

    def test_add_remove_domain(self, session_browser):
        """Test adding and removing a domain."""
        functional.app_enable(session_browser, 'ejabberd')
        functional.app_enable(session_browser, 'avahi')
        functional.set_hostname(session_browser, 'freedombox')
        _enable_domain(session_browser, 'freedombox.local')

        _disable_domain(session_browser, 'freedombox.local')
        assert not _is_domain_enabled(session_browser, 'freedombox.local')
        assert functional.service_is_running(session_browser, 'ejabberd')

        _enable_domain(session_browser, 'freedombox.local')
        assert _is_domain_enabled(session_browser, 'freedombox.local')
        assert functional.service_is_running(session_browser, 'ejabberd')

    def test_message_archive_management(self, session_browser):
        """Test enabling message archive management."""
        functional.app_enable(session_browser, 'ejabberd')
        _enable_message_archive_management(session_browser)
        assert functional.service_is_running(session_browser, 'ejabberd')

        _disable_message_archive_management(session_browser)
        assert functional.service_is_running(session_browser, 'ejabberd')

    @pytest.mark.backups
    def test_backup_restore(self, session_browser):
        """Test backup and restore of app data."""
        functional.app_enable(session_browser, 'ejabberd')
        functional.app_enable(session_browser, 'jsxc')
        _jsxc_add_contact(session_browser)
        functional.backup_create(session_browser, 'ejabberd', 'test_ejabberd')

        _jsxc_delete_contact(session_browser)
        functional.backup_restore(session_browser, 'ejabberd', 'test_ejabberd')

        _jsxc_assert_has_contact(session_browser)


def _enable_domain(browser, domain):
    """Add domain name to Ejabberd configuration."""
    functional.nav_to_module(browser, 'ejabberd')
    checkbox = browser.find_by_value(domain).first
    checkbox.check()
    functional.submit(browser, form_class='form-configuration')


def _disable_domain(browser, domain):
    """Remove domain name from Ejabberd configuration."""
    functional.nav_to_module(browser, 'ejabberd')
    checkbox = browser.find_by_value(domain).first
    checkbox.uncheck()
    functional.submit(browser, form_class='form-configuration')


def _is_domain_enabled(browser, domain):
    """Return whether the domain name is enabled."""
    functional.nav_to_module(browser, 'ejabberd')
    checkbox = browser.find_by_value(domain).first
    return checkbox.checked


def _enable_message_archive_management(browser):
    """Enable Message Archive Management in Ejabberd."""
    functional.nav_to_module(browser, 'ejabberd')
    functional.change_checkbox_status(browser, 'ejabberd', 'id_MAM_enabled',
                                      'enabled')


def _disable_message_archive_management(browser):
    """Enable Message Archive Management in Ejabberd."""
    functional.nav_to_module(browser, 'ejabberd')
    functional.change_checkbox_status(browser, 'ejabberd', 'id_MAM_enabled',
                                      'disabled')


def _is_jsxc_buddy_list_loaded(browser):
    """Return whether the jsxc buddy list has been loaded."""
    if browser.find_by_text('new contact'):
        # no contacts
        return True

    buddy_list = browser.find_by_id('jsxc_buddylist').first
    contacts = buddy_list.find_by_css('.jsxc_rosteritem')

    return len(contacts) > 0


def _jsxc_login(browser):
    """Login to JSXC."""
    username = functional.config['DEFAULT']['username']
    password = functional.config['DEFAULT']['password']
    functional.visit(browser, '/plinth/apps/jsxc/jsxc/')
    assert functional.eventually(browser.find_by_text,
                                 ['BOSH Server reachable.'])
    if browser.find_by_text('relogin'):
        browser.reload()

    browser.find_by_id('jsxc-username').fill(username)
    browser.find_by_id('jsxc-password').fill(password)
    browser.find_by_id('jsxc-submit').click()
    assert functional.eventually(browser.find_by_css,
                                 ['#jsxc_roster.jsxc_state_shown'])


def _jsxc_add_contact(browser):
    """Add a contact to JSXC user's roster."""
    functional.set_domain_name(browser, 'localhost')
    functional.install(browser, 'jsxc')
    _jsxc_login(browser)
    functional.eventually(_is_jsxc_buddy_list_loaded, args=[browser])
    new = browser.find_by_text('new contact')
    if new:  # roster is empty
        new.first.click()
        browser.find_by_id('jsxc_username').fill('alice@localhost')
        browser.find_by_text('Add').first.click()
        assert functional.eventually(browser.find_by_text, ['alice@localhost'])


def _jsxc_delete_contact(browser):
    """Delete the contact from JSXC user's roster."""
    _jsxc_login(browser)

    # noqa, pylint: disable=unnecessary-lambda
    functional.eventually(browser.find_by_css, ['div.jsxc_more'])
    browser.find_by_css('div.jsxc_more').first.click()
    functional.eventually(browser.find_by_text, ['delete contact'])
    browser.find_by_text('delete contact').first.click()
    functional.eventually(browser.find_by_text, ['Remove'])
    browser.find_by_text('Remove').first.click()


def _jsxc_assert_has_contact(browser):
    """Check whether the contact is in JSXC user's roster."""
    _jsxc_login(browser)
    assert functional.eventually(browser.find_by_text, ['alice@localhost'])



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Nextcloud configuration form."""

from django import forms
from django.utils.translation import gettext_lazy as _


def _get_phone_regions():
    """Return choice field choices for phone regions."""
    try:
        from iso3166 import countries  # type: ignore
        phone_regions = [(country.alpha2, country.name)
                         for country in countries]
        phone_regions = sorted(phone_regions)
    except ImportError:
        # Allow users to set a non-empty value
        phone_regions = [('US', 'United States of America')]

    return [('', _('Not set'))] + phone_regions


class NextcloudForm(forms.Form):
    """Nextcloud configuration form."""

    override_domain = forms.CharField(
        label=_('Override domain'), required=False, help_text=_(
            'Set to the domain or IP address that Nextcloud should be forced '
            'to generate URLs with. Should not be needed if a valid domain is '
            'used to access Nextcloud. Examples: "myfreedombox.example.org" '
            'or "example.onion".'))

    admin_password = forms.CharField(
        label=_('Administrator password'), help_text=_(
            'Optional. Set a new password for Nextcloud\'s administrator '
            'account (nextcloud-admin). The password cannot be a common one '
            'and the minimum required length is <strong>10 characters'
            '</strong>. Leave this field blank to keep the current password.'),
        required=False, widget=forms.PasswordInput, min_length=10)

    default_phone_region = forms.ChoiceField(
        label=_('Default phone region'), required=False,
        help_text=_('The default phone region is required to validate phone '
                    'numbers in the profile settings without a country code.'),
        choices=_get_phone_regions)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
FreedomBox app for Kiwix content server.
"""

from django.utils.translation import gettext_lazy as _

from plinth import app as app_module
from plinth import frontpage, menu, package
from plinth.config import DropinConfigs
from plinth.daemon import Daemon
from plinth.modules.apache.components import Webserver
from plinth.modules.backups.components import BackupRestore
from plinth.modules.firewall.components import (Firewall,
                                                FirewallLocalProtection)
from plinth.modules.users.components import UsersAndGroups

from . import manifest, privileged

_description = [
    _('Kiwix is an offline reader for web content. It is software intended '
      'to make Wikipedia available without using the internet, but it is '
      'potentially suitable for all HTML content. Kiwix packages are in the '
      'ZIM file format.'),
    _('''Kiwix can host various kinds of content:
    <ul>
      <li>Offline versions of websites: Wikimedia projects, Stack Exchange</li>
      <li>Video content: Khan Academy, TED Talks, Crash Course</li>
      <li>Educational materials: PHET, TED Ed, Vikidia</li>
      <li>eBooks: Project Gutenberg</li>
      <li>Magazines: Low-tech Magazine</li>
    </ul>'''),
    _('You can <a href="https://library.kiwix.org" target="_blank" '
      'rel="noopener noreferrer">download</a> content packages from the Kiwix '
      'project or <a href="https://openzim.org/wiki/Build_your_ZIM_file" '
      'target="_blank" rel="noopener noreferrer">create</a> your own.'),
]


class KiwixApp(app_module.App):
    """FreedomBox app for Kiwix."""

    app_id = 'kiwix'

    _version = 1

    DAEMON = 'kiwix-server-freedombox'

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()

        groups = {'kiwix': _('Manage Kiwix content server')}

        info = app_module.Info(
            app_id=self.app_id, version=self._version, name=_('Kiwix'),
            icon_filename='kiwix', short_description=_('Offline Wikipedia'),
            description=_description, manual_page='Kiwix',
            clients=manifest.clients,
            donation_url='https://www.kiwix.org/en/support-us/')
        self.add(info)

        menu_item = menu.Menu('menu-kiwix', info.name, info.short_description,
                              info.icon_filename, 'kiwix:index',
                              parent_url_name='apps')
        self.add(menu_item)

        shortcut = frontpage.Shortcut('shortcut-kiwix', info.name,
                                      short_description=info.short_description,
                                      icon=info.icon_filename, url='/kiwix',
                                      clients=info.clients,
                                      login_required=False,
                                      allowed_groups=list(groups))
        self.add(shortcut)

        packages = package.Packages('packages-kiwix', ['kiwix-tools'])
        self.add(packages)

        dropin_configs = DropinConfigs('dropin-configs-kiwix', [
            '/etc/apache2/conf-available/kiwix-freedombox.conf',
        ])
        self.add(dropin_configs)

        firewall = Firewall('firewall-kiwix', info.name,
                            ports=['http', 'https'], is_external=True)
        self.add(firewall)

        firewall_local_protection = FirewallLocalProtection(
            'firewall-local-protection-kiwix', ['4201'])
        self.add(firewall_local_protection)

        webserver = Webserver('webserver-kiwix', 'kiwix-freedombox',
                              urls=['https://{host}/kiwix'])
        self.add(webserver)

        daemon = Daemon('daemon-kiwix', self.DAEMON,
                        listen_ports=[(4201, 'tcp4')])
        self.add(daemon)

        users_and_groups = UsersAndGroups('users-and-groups-kiwix',
                                          reserved_usernames=['kiwix'],
                                          groups=groups)
        self.add(users_and_groups)

        backup_restore = BackupRestore('backup-restore-kiwix',
                                       **manifest.backup)
        self.add(backup_restore)

    def setup(self, old_version=None):
        """Install and configure the app."""
        super().setup(old_version)
        if not old_version:
            self.enable()

    def uninstall(self):
        """De-configure and uninstall the app."""
        super().uninstall()
        privileged.uninstall()


def validate_file_name(file_name: str):
    """Check if the content package file has a valid extension."""
    if not file_name.endswith('.zim'):
        raise ValueError(f'Expected a ZIM file. Found {file_name}')



#
# This file is part of FreedomBox.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Basic settings for Django web framework.

During initialization of the service, these settings are overridden before
Django is initialized. However, this file has been written in this format to
let Django initialization easier in other situations such as test cases,
documentation generation, debugging etc.

See: https://docs.djangoproject.com/en/dev/ref/settings/

"""

ALLOWED_HOSTS = ['*']

_pwd = 'django.contrib.auth.password_validation'

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': '{}.UserAttributeSimilarityValidator'.format(_pwd),
    },
    {
        'NAME': '{}.MinimumLengthValidator'.format(_pwd),
        'OPTIONS': {
            'min_length': 8,
        }
    },
    {
        'NAME': '{}.CommonPasswordValidator'.format(_pwd),
    },
    {
        'NAME': '{}.NumericPasswordValidator'.format(_pwd),
    },
]

AUTHENTICATION_BACKENDS = [
    # AxesStandaloneBackend should be the first backend. If the user is locked
    # out due to too many attempts, the backend denies further attempts untill
    # unlocked by a CAPTCHA form.
    'axes.backends.AxesStandaloneBackend',

    # Django ModelBackend is the default authentication backend.
    'django.contrib.auth.backends.ModelBackend',
]

AXES_LOCKOUT_URL = 'locked/'

AXES_RESET_ON_SUCCESS = True

AXES_VERBOSE = False

CACHES = {
    'default': {
        'BACKEND': 'django.core.cache.backends.dummy.DummyCache'
    }
}

CAPTCHA_FONT_PATH = ['/usr/share/fonts/truetype/ttf-bitstream-vera/Vera.ttf']

CAPTCHA_LENGTH = 6

CAPTCHA_FLITE_PATH = '/usr/bin/flite'

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'OPTIONS': {
            'timeout': 30
        },
        # Overridden based on the configuration key store_file
        'NAME': '/var/lib/plinth/plinth.sqlite3'
    }
}

# Overridden based on command line argument --develop
DEBUG = False

# This is already the default (Django 3.2), however, setting it explicitly
# seems to avoid a warning while running 'django-admin makemigrations'.
DEFAULT_AUTO_FIELD = 'django.db.models.AutoField'

# Overridden based on the configuration key server_dir
FORCE_SCRIPT_NAME = '/plinth'

# FreedomBox apps are appended to this list
INSTALLED_APPS = [
    'axes',
    'captcha',
    'bootstrapform',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.messages',
    'django.contrib.sessions',
    'stronghold',
    'plinth',
]

# Overridden based on configuration key use_x_forwarded_host
IPWARE_META_PRECEDENCE_ORDER = ('REMOTE_ADDR', )

# Overridden by get_languages()
LANGUAGES = [('en', 'English')]

# Overridden by log configuration in log.py
LOGGING = {'version': 1}

LOGIN_URL = 'users:login'

LOGIN_REDIRECT_URL = 'index'

# Overridden before initialization
MESSAGE_TAGS: dict = {}

MIDDLEWARE = (
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.locale.LocaleMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'plinth.middleware.CommonErrorMiddleware',
    'stronghold.middleware.LoginRequiredMiddleware',
    'plinth.middleware.AdminRequiredMiddleware',
    'plinth.middleware.FirstSetupMiddleware',
    'plinth.modules.first_boot.middleware.FirstBootMiddleware',
    'plinth.middleware.SetupMiddleware',

    # AxesMiddleware should be the last middleware. It only formats user
    # lockout messages and renders Axes lockout responses on failed user
    # authentication attempts from login views.
    'axes.middleware.AxesMiddleware',
)

PASSWORD_HASHERS = [
    'plinth.hashers.Argon2PasswordHasherLowMemory',
    'django.contrib.auth.hashers.PBKDF2PasswordHasher',
    'django.contrib.auth.hashers.PBKDF2SHA1PasswordHasher',
    'django.contrib.auth.hashers.BCryptSHA256PasswordHasher',
]

ROOT_URLCONF = 'plinth.urls'

SECURE_CONTENT_TYPE_NOSNIFF = True

# Overridden based configuration key secure_proxy_ssl_header
SECURE_PROXY_SSL_HEADER = None

SESSION_ENGINE = 'django.contrib.sessions.backends.file'

SESSION_FILE_PATH = '/var/lib/plinth/sessions'

# Overridden based on configuration key server_dir
STATIC_URL = '/plinth/static/'

# STRONGHOLD_PUBLIC_URLS=(r'^captcha/', )

STRONGHOLD_PUBLIC_NAMED_URLS = (
    'captcha-image',
    'captcha-image-2x',
    'captcha-audio',
    'captcha-refresh',
)

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.contrib.auth.context_processors.auth',
                'django.template.context_processors.debug',
                'django.template.context_processors.i18n',
                'django.template.context_processors.media',
                'django.template.context_processors.request',
                'django.template.context_processors.static',
                'django.template.context_processors.tz',
                'django.contrib.messages.context_processors.messages',
                'plinth.context_processors.common',
            ],
        },
    },
]

TIME_ZONE = 'UTC'

USE_L10N = True

USE_TZ = True

# Overridden by configuration setting use_x_forwarded_host
USE_X_FORWARDED_HOST = False



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
App component to introduce a new domain type.
"""

from typing import ClassVar

from django.utils.translation import gettext_lazy as _

from plinth import app

_SERVICES = {
    '__all__': {
        'display_name': _('All'),
        'port': None
    },
    'http': {
        'display_name': _('All web apps'),
        'port': 80
    },
    'https': {
        'display_name': _('All web apps'),
        'port': 443
    },
    'ssh': {
        'display_name': _('Secure Shell'),
        'port': 22
    },
}


class DomainType(app.FollowerComponent):
    """Component to create a new type of domain.

    It serves the primary purpose of showing a entry in the domain names page.
    This will allow users to discover the new type of domain and use the
    providing app to create that type of domain.

    Similar to a menu entry, domain type information is available to the user
    even when the corresponding app is disabled.

    """

    _all: ClassVar[dict[str, 'DomainType']] = {}

    def __init__(self, component_id, display_name, configuration_url,
                 can_have_certificate=True):
        """Initialize the domain type component.

        component_id should be a unique ID across all components of an app and
        across all components. This will also act as the 'type' parameter for
        each created domain.

        display_name is the type of domain displayed to the user in the
        interface.

        configuration_url is the Django URL to which a user is redirected to in
        order to create or manage a domain of this type.

        can_have_certificate indicates if this type of domain can have a TLS
        certificate that can be validated by a typical browser.

        """
        super().__init__(component_id)

        self.display_name = display_name
        self.configuration_url = configuration_url
        self.can_have_certificate = can_have_certificate

        self._all[component_id] = self

    @classmethod
    def get(cls, component_id):
        """Return a component of given ID."""
        return cls._all[component_id]

    @classmethod
    def list(cls):
        """Return a list of all domain types."""
        return dict(cls._all)


class DomainName(app.FollowerComponent):
    """Component to represent a domain name and its properties.

    Each domain name is necessarily attached to a domain type component that
    must be created prior to creating the domain name.

    When an application providing or managing a domain name is disabled, the
    corresponding domain name should become unavailable for others apps and
    they must de-configure the domain name from app configuration. This is the
    primary reason for making a domain name available as a component.

    """
    _all: ClassVar[dict[str, 'DomainName']] = {}

    def __init__(self, component_id, name, domain_type, services):
        """Initialize a domain name.

        component_id should be a unique ID across all components of an app and
        across all components. The value is typically 'domain-{app}-{domain}'.
        This ensures that if the same domain is managed by multiple apps, it is
        available as multiple instances. When one instance is removed, say by
        disabling that app, the other instance will still provide that domain.

        name is the domain name that is being represented by the component.
        This should be fully qualified domain name.

        domain_type should be a string representing the type of the domain.
        This is the component ID of the domain type earlier registered by the
        app that is creating the domain name component.

        services is a list of string identifiers for services potentially
        provided by the domain. For example, 'ssh' for secure shell service
        provided on port 22. It is used for showing information to the user and
        to retrieve a list of a domains that an app may use.

        A service value can also be an integer in which case it will be
        converted to a string by looking up a list of known services. This
        process is not perfect and may cause problems when filtering domains
        that could potentially provide a service.

        The most common value of a services parameter is the string '__all__'
        indicating that the domain can potentially provide any service without
        limitations due to the nature of the domain name.

        """
        super().__init__(component_id)

        self.name = name
        self.domain_type = DomainType.get(domain_type)
        self._services = DomainName._normalize_services(services)

        self._all[component_id] = self

    @property
    def services(self):
        """Read-only property to get the list of services."""
        return self._services

    @staticmethod
    def _normalize_services(services):
        """If ports numbers are provided convert them to service IDs."""
        if services == '__all__':
            return services

        return [DomainName._normalize_service(service) for service in services]

    @staticmethod
    def _normalize_service(service):
        """Return the service ID for a given port number.

        XXX: Eliminate this and use a generalized approach eventually.

        """
        if isinstance(service, str):
            return service

        if not isinstance(service, int):
            raise ValueError('Invalid service')

        for service_id, description in _SERVICES.items():
            if description['port'] == service:
                return service_id

        return str(service)

    def get_readable_services(self):
        """Return list of unique service strings that can be shown to user."""
        services = self.services
        if self.services == '__all__':
            services = [services]

        return {
            _SERVICES.get(service, {'display_name': service})['display_name']
            for service in services
        }

    def has_service(self, service):
        """Return whether a service is available for this domain name."""
        return (service is None or self.services == '__all__'
                or service in self.services)

    def remove(self):
        """Remove the domain name from global list of domains.

        It is acceptable to call remove() multiple times.

        """
        try:
            del self._all[self.component_id]
        except KeyError:
            pass

    @classmethod
    def get(cls, component_id):
        """Return the domain name object given name and app."""
        return cls._all[component_id]

    @classmethod
    def list(cls, filter_for_service=None):
        """Return list of domains."""
        return [
            domain for domain in cls._all.values()
            if domain.has_service(filter_for_service)
        ]

    @classmethod
    def list_names(cls, filter_for_service=None):
        """Return a set of unique domain names.

        Multiple different components may provide the same domain name. This
        method could be used to retrieve a list of all domain names without
        duplication.

        """
        return {
            domain.name
            for domain in cls._all.values()
            if domain.has_service(filter_for_service)
        }



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
URLs for the disks module.
"""

from django.urls import re_path

from . import views

urlpatterns = [
    re_path(r'^sys/storage/$', views.StorageAppView.as_view(), name='index'),
    re_path(r'^sys/storage/expand$', views.expand, name='expand'),
    re_path(r'^sys/storage/eject/(?P<device_path>[A-Za-z0-9%_.\-~]+)/$',
            views.eject, name='eject')
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Test module for menus.
"""

import random

import pytest
from django.http import HttpRequest
from django.urls import reverse

from plinth import menu as menu_module
from plinth.menu import Menu

URL_TEMPLATE = '/test/{}/{}/{}/'

# Test helper methods


def build_menu(size=5):
    """Build a menu with the specified number of items."""
    random.seed()

    menu = Menu('menu-index', url_name='index')

    for index in range(1, size + 1):
        kwargs = {
            'component_id': 'menu-test-' + str(index),
            'name': 'Name' + str(index),
            'short_description': 'ShortDescription' + str(index),
            'icon': 'Icon' + str(index),
            'url_name': 'test',
            'url_kwargs': {
                'a': index,
                'b': index,
                'c': index
            },
            'parent_url_name': 'index',
            'order': random.randint(0, 1000),
        }

        Menu(**kwargs)

    return menu


@pytest.fixture(name='empty_menus', autouse=True)
def fixture_empty_menus():
    """Remove all menu entries before starting a test."""
    Menu._all_menus = set()


def test_init(rf):
    """Verify that main_menu and essential items are created."""
    menu_module.init()
    main_menu = menu_module.main_menu
    assert isinstance(main_menu, Menu)

    apps_menu = main_menu.active_item(rf.get('/apps/foo/'))
    assert apps_menu.icon == 'fa-download'
    assert str(apps_menu.url) == '/apps/'

    system_menu = main_menu.active_item(rf.get('/sys/bar/'))
    assert system_menu.icon == 'fa-cog'
    assert str(system_menu.url) == '/sys/'


def test_menu_creation_without_arguments():
    """Verify the Menu state without initialization parameters."""
    with pytest.raises(ValueError):
        Menu('menu-test')

    menu = Menu('menu-index', url_name='index')
    assert menu.component_id == 'menu-index'
    assert menu.name is None
    assert menu.short_description is None
    assert menu.icon is None
    assert menu.url == '/'
    assert menu.order == 50
    assert not menu.advanced
    assert not menu.items


def test_menu_creation_with_arguments():
    """Verify the Menu state with initialization parameters."""
    expected_name = 'Name'
    expected_short_description = 'ShortDescription'
    expected_icon = 'Icon'
    url_name = 'test'
    url_kwargs = {'a': 1, 'b': 2, 'c': 3}
    expected_url = reverse(url_name, kwargs=url_kwargs)
    expected_order = 42
    parent_menu = Menu('menu-index', url_name='index')
    menu = Menu('menu-test', expected_name, expected_short_description,
                expected_icon, url_name, url_kwargs=url_kwargs,
                parent_url_name='index', order=expected_order, advanced=True)

    assert menu.parent_url_name == 'index'
    assert len(parent_menu.items) == 1
    assert parent_menu.items[0] == menu
    assert expected_name == menu.name
    assert expected_short_description == menu.short_description
    assert expected_icon == menu.icon
    assert expected_url == menu.url
    assert expected_order == menu.order
    assert menu.advanced
    assert url_name == menu.url_name
    assert menu.url_args is None
    assert url_kwargs == menu.url_kwargs
    assert not menu.items


def test_sort_items():
    """Verify that menu items are sorted correctly."""
    size = 1000
    menu = build_menu(size)

    for index in range(0, 200):
        menu.items[index].order = 100

    # Verify that the order of every item is equal to or greater
    # than the order of the item preceding it and if the order is
    # the same, the names are considered.
    items = menu.sorted_items()
    for index in range(1, size):
        assert items[index].order >= items[index - 1].order
        if items[index].order == items[index - 1].order:
            assert items[index].name >= items[index - 1].name


def test_active_item():
    """Verify that an active menu item can be correctly retrieved."""
    menu = build_menu()

    for index in range(1, 8):
        request = HttpRequest()
        request.path = URL_TEMPLATE.format(index, index, index)
        item = menu.active_item(request)
        if index <= 5:
            assert request.path == item.url
        else:
            assert item is None


def test_active_item_when_inside_subpath():
    """Verify that the current URL could be a sub-path of a menu item."""
    menu = build_menu()
    expected_url = URL_TEMPLATE.format(1, 1, 1)
    request = HttpRequest()
    request.path = expected_url + 'd/e/f/'
    item = menu.active_item(request)
    assert expected_url == item.url



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app to configure PageKite."""

from django.utils.translation import gettext_lazy as _

from plinth import app as app_module
from plinth import cfg, menu
from plinth.daemon import Daemon
from plinth.modules.backups.components import BackupRestore
from plinth.modules.names.components import DomainType
from plinth.package import Packages
from plinth.privileged import service as service_privileged
from plinth.utils import format_lazy

from . import manifest, utils

_description = [
    format_lazy(
        _('PageKite is a system for exposing {box_name} services when '
          'you don\'t have a direct connection to the Internet. You only '
          'need this if your {box_name} services are unreachable from '
          'the rest of the Internet. This includes the following '
          'situations:'), box_name=_(cfg.box_name)),
    format_lazy(_('{box_name} is behind a restricted firewall.'),
                box_name=_(cfg.box_name)),
    format_lazy(
        _('{box_name} is connected to a (wireless) router which you '
          'don\'t control.'), box_name=_(cfg.box_name)),
    _('Your ISP does not provide you an external IP address and '
      'instead provides Internet connection through NAT.'),
    _('Your ISP does not provide you a static IP address and your IP '
      'address changes every time you connect to Internet.'),
    _('Your ISP limits incoming connections.'),
    format_lazy(
        _('PageKite works around NAT, firewalls and IP address limitations '
          'by using a combination of tunnels and reverse proxies. You can '
          'use any pagekite service provider, for example '
          '<a href="https://pagekite.net">pagekite.net</a>.  In the future it '
          'might be possible to use your buddy\'s {box_name} for this.'),
        box_name=_(cfg.box_name))
]


class PagekiteApp(app_module.App):
    """FreedomBox app for Pagekite."""

    app_id = 'pagekite'

    DAEMON = 'pagekite'

    _version = 2

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()

        info = app_module.Info(
            app_id=self.app_id, version=self._version, depends=['names'],
            name=_('PageKite'), icon='fa-flag',
            short_description=_('Public Visibility'), description=_description,
            manual_page='PageKite',
            donation_url='https://pagekite.net/support/faq/#donate')
        self.add(info)

        menu_item = menu.Menu('menu-pagekite', info.name,
                              info.short_description, info.icon,
                              'pagekite:index',
                              parent_url_name='system:visibility', order=40)
        self.add(menu_item)

        packages = Packages('packages-pagekite', ['pagekite'])
        self.add(packages)

        domain_type = DomainType('domain-type-pagekite', _('PageKite Domain'),
                                 'pagekite:index', can_have_certificate=True)
        self.add(domain_type)

        daemon = Daemon('daemon-pagekite', self.DAEMON)
        self.add(daemon)

        backup_restore = BackupRestore('backup-restore-pagekite',
                                       **manifest.backup)
        self.add(backup_restore)

    def post_init(self):
        """Perform post initialization operations."""
        # Register kite name with Name Services module.
        if not self.needs_setup() and self.is_enabled():
            utils.update_names_module(is_enabled=True)

    def enable(self):
        """Send domain signals after enabling the app."""
        super().enable()
        utils.update_names_module(is_enabled=True)

    def disable(self):
        """Send domain signals before disabling the app."""
        utils.update_names_module(is_enabled=False)
        super().disable()

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)
        if not old_version:
            self.enable()

        if old_version == 1:
            service_privileged.try_restart(PagekiteApp.DAEMON)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Test module for package module.
"""

import unittest
from unittest.mock import Mock, call, patch

import pytest

from plinth.app import App
from plinth.diagnostic_check import DiagnosticCheck, Result
from plinth.errors import MissingPackageError
from plinth.package import Package, Packages, packages_installed


@pytest.fixture(autouse=True)
def fixture_clean_apps():
    """Fixture to ensure clean set of global apps."""
    App._all_apps = {}


class TestPackageExpressions(unittest.TestCase):

    def test_package(self):
        """Test resolving package names."""
        package = Package('python3')
        assert package.possible() == ['python3']
        assert package.actual() == 'python3'

        package = Package('unknown-package')
        assert package.possible() == ['unknown-package']
        self.assertRaises(MissingPackageError, package.actual)

    def test_package_or_expression(self):
        """Test resolving package OR expressions."""
        expression = Package('python3') | Package('unknown-package')
        assert expression.possible() == ['python3', 'unknown-package']
        assert expression.actual() == 'python3'

        expression = Package('unknown-package') | Package('python3')
        assert expression.possible() == ['unknown-package', 'python3']
        assert expression.actual() == 'python3'

        # When both packages are available, prefer the first.
        expression = Package('bash') | Package('dash')
        assert expression.possible() == ['bash', 'dash']
        assert expression.actual() == 'bash'

        expression = Package('unknown-package') | Package(
            'another-unknown-package')
        assert expression.possible() == [
            'unknown-package', 'another-unknown-package'
        ]
        self.assertRaises(MissingPackageError, expression.actual)


def test_packages_init():
    """Test initialization of packages component."""
    component = Packages('test-component', ['foo', 'bar'])
    assert component.possible_packages == ['foo', 'bar']
    assert component.component_id == 'test-component'
    assert not component.skip_recommends
    assert component.conflicts == []
    assert component.conflicts_action is None

    with pytest.raises(ValueError):
        Packages(None, [])

    component = Packages('test-component', [], skip_recommends=True,
                         conflicts=['conflict1', 'conflict2'],
                         conflicts_action=Packages.ConflictsAction.IGNORE)
    assert component.possible_packages == []
    assert component.skip_recommends
    assert component.conflicts == ['conflict1', 'conflict2']
    assert component.conflicts_action == Packages.ConflictsAction.IGNORE


def test_packages_get_actual_packages():
    """Test resolving of package expressions to actual packages."""
    component = Packages('test-component', ['python3'])
    assert component.get_actual_packages() == ['python3']

    component = Packages('test-component',
                         [Package('unknown-package') | Package('python3')])
    assert component.get_actual_packages() == ['python3']

    component = Packages('test-component', [], skip_recommends=True,
                         conflicts=['conflict1', 'conflict2'],
                         conflicts_action=Packages.ConflictsAction.IGNORE)
    assert component.get_actual_packages() == []


@patch('plinth.package.install')
def test_packages_setup(install):
    """Test setting up packages component."""

    class TestApp(App):
        """Test app"""
        app_id = 'test-app'

    component = Packages('test-component', ['python3', 'bash'])
    app = TestApp()
    app.add(component)
    app.setup(old_version=3)
    install.assert_has_calls(
        [call(['python3', 'bash'], skip_recommends=False)])

    component = Packages('test-component', ['bash', 'perl'],
                         skip_recommends=True)
    app = TestApp()
    app.add(component)
    app.setup(old_version=3)
    install.assert_has_calls([call(['bash', 'perl'], skip_recommends=True)])

    component = Packages('test-component',
                         [Package('python3') | Package('unknown-package')])
    app = TestApp()
    app.add(component)
    app.setup(old_version=3)
    install.assert_has_calls([call(['python3'], skip_recommends=False)])


@patch('plinth.package.packages_installed')
@patch('plinth.package.uninstall')
@patch('plinth.package.install')
def test_packages_setup_with_conflicts(install, uninstall, packages_installed):
    """Test setting up packages with conflicts."""
    packages_installed.return_value = ['exim4-base']

    component = Packages('test-component', ['bash'], conflicts=['exim4-base'],
                         conflicts_action=Packages.ConflictsAction.REMOVE)
    component.setup(old_version=0)
    uninstall.assert_has_calls([call(['exim4-base'], purge=False)])
    install.assert_has_calls([call(['bash'], skip_recommends=False)])

    uninstall.reset_mock()
    install.reset_mock()
    component = Packages('test-component', ['bash'], conflicts=['exim4-base'])
    component.setup(old_version=0)
    uninstall.assert_not_called()
    install.assert_has_calls([call(['bash'], skip_recommends=False)])

    uninstall.reset_mock()
    install.reset_mock()
    component = Packages('test-component', ['bash'],
                         conflicts=['exim4-base', 'not-installed-package'],
                         conflicts_action=Packages.ConflictsAction.IGNORE)
    component.setup(old_version=0)
    uninstall.assert_not_called()
    install.assert_has_calls([call(['bash'], skip_recommends=False)])


@patch('plinth.package.refresh_package_lists')
@patch('plinth.package.uninstall')
def test_packages_uninstall(uninstall, _refresh_package_lists):
    """Test uninstalling packages component."""

    class TestApp(App):
        """Test app"""
        app_id = 'test-app'

    component = Packages('test-component', ['python3', 'bash'])
    app = TestApp()
    app.add(component)
    app.uninstall()
    uninstall.assert_has_calls([call(['python3', 'bash'], purge=True)])


@patch('plinth.package.refresh_package_lists')
@patch('plinth.package.uninstall')
@patch('apt.Cache')
def test_packages_uninstall_exclusion(cache, uninstall,
                                      _refresh_package_lists):
    """Test excluding packages from other installed apps when uninstalling."""

    def _get_mock_package(installed_version='1.0', dependencies=None,
                          recommends=None):
        mock_dependencies = []
        for or_dependencies in (dependencies or []):
            mock_or_dependency = Mock(or_dependencies=[])
            mock_dependencies.append(mock_or_dependency)
            for dependency in or_dependencies:
                mock = Mock()
                mock.name = dependency
                mock_or_dependency.or_dependencies.append(mock)

        mock_recommends = []
        for or_dependencies in (recommends or []):
            mock_or_dependency = Mock(or_dependencies=[])
            mock_recommends.append(mock_or_dependency)
            for dependency in or_dependencies:
                mock = Mock()
                mock.name = dependency
                mock_or_dependency.or_dependencies.append(mock)

        mock = Mock(
            version=installed_version or '1.0',
            is_installed=bool(installed_version),
            installed=Mock(dependencies=mock_dependencies,
                           recommends=mock_recommends))
        return mock

    package2 = _get_mock_package('4.0', [['dep1', 'dep2'], ['dep3'], ['dep4']],
                                 [['dep5']])
    cache.return_value = {
        'package11': _get_mock_package('2.0'),
        'package12': _get_mock_package(None),
        'package2': package2,
        'package3': _get_mock_package('5.0', ['unknown-dep1']),
        'dep1': _get_mock_package('6.0'),
        'dep2': _get_mock_package('6.1'),
        'dep3': _get_mock_package('6.2'),
        'dep4': _get_mock_package(None),
        'dep5': _get_mock_package('6.4'),
        'dep6': _get_mock_package('6.5'),
    }

    class TestApp1(App):
        """Test app."""
        app_id = 'test-app1'

        def __init__(self):
            super().__init__()
            component = Packages('test-component11', [
                'package11', 'package2', 'package3', 'dep1', 'dep2', 'dep3',
                'dep4', 'dep6'
            ])
            self.add(component)

            component = Packages('test-component12',
                                 ['package12', 'package2', 'package3'])
            self.add(component)

    class TestApp2(App):
        """Test app."""
        app_id = 'test-app2'

        def __init__(self):
            super().__init__()
            component = Packages('test-component2', ['package2'])
            self.add(component)

        def get_setup_state(self):
            return App.SetupState.UP_TO_DATE

    class TestApp3(App):
        """Test app."""
        app_id = 'test-app3'

        def __init__(self):
            super().__init__()
            component = Packages('test-component3', ['package3'])
            self.add(component)

        def get_setup_state(self):
            return App.SetupState.NEEDS_SETUP

    app1 = TestApp1()
    TestApp2()
    TestApp3()
    app1.uninstall()
    uninstall.assert_has_calls([
        call(['package11', 'package3', 'dep6'], purge=True),
        call(['package12', 'package3'], purge=True)
    ])


@patch('apt.Cache')
def test_diagnose(cache):
    """Test checking for latest version of the package."""
    cache.return_value = {
        'package2': Mock(candidate=Mock(version='2.0', is_installed=True)),
        'package3': Mock(candidate=Mock(version='3.0', is_installed=False)),
        'package7': Mock(candidate=Mock(version='4.0', is_installed=True)),
    }
    component = Packages('test-component', [
        'package1', 'package2', 'package3',
        Package('package4') | Package('package5'),
        Package('package6') | Package('package7')
    ])
    results = component.diagnose()
    assert results == [
        DiagnosticCheck(
            'package-available-package1',
            'Package {package_expression} is not available for install',
            Result.FAILED, {'package_expression': 'package1'},
            'test-component'),
        DiagnosticCheck(
            'package-latest-package2',
            'Package {package_name} is the latest version ({latest_version})',
            Result.PASSED, {
                'package_name': 'package2',
                'latest_version': '2.0'
            }, 'test-component'),
        DiagnosticCheck(
            'package-latest-package3',
            'Package {package_name} is the latest version ({latest_version})',
            Result.WARNING, {
                'package_name': 'package3',
                'latest_version': '3.0'
            }, 'test-component'),
        DiagnosticCheck(
            'package-available-package4 | package5',
            'Package {package_expression} is not available for install',
            Result.FAILED, {'package_expression': 'package4 | package5'},
            'test-component'),
        DiagnosticCheck(
            'package-latest-package7',
            'Package {package_name} is the latest version ({latest_version})',
            Result.PASSED, {
                'package_name': 'package7',
                'latest_version': '4.0'
            }, 'test-component'),
    ]


@patch('plinth.package.packages_installed')
def test_packages_find_conflicts(packages_installed_):
    """Test finding conflicts."""
    packages_installed_.return_value = []
    component = Packages('test-component', ['package3', 'package4'])
    assert component.find_conflicts() is None

    packages_installed_.return_value = []
    component = Packages('test-component', ['package3', 'package4'],
                         conflicts=['package5', 'package6'],
                         conflicts_action=Packages.ConflictsAction.IGNORE)
    assert component.find_conflicts() == []

    packages_installed_.return_value = ['package1', 'package2']
    component = Packages('test-component', ['package3', 'package4'],
                         conflicts=['package1', 'package2'],
                         conflicts_action=Packages.ConflictsAction.IGNORE)
    assert component.find_conflicts() == ['package1', 'package2']


@patch('apt.Cache')
@patch('pathlib.Path')
def test_packages_has_unavailable_packages(path_class, cache):
    """Test checking for unavailable packages."""
    path = Mock()
    path_class.return_value = path
    path.iterdir.return_value = [Mock()]

    component = Packages('test-component', ['package1', 'package2'])
    assert component.has_unavailable_packages() is None

    path.iterdir.return_value = [Mock(), Mock()]
    cache.return_value = ['package1', 'package2']
    assert not component.has_unavailable_packages()

    cache.return_value = ['package1']
    assert component.has_unavailable_packages()


def test_packages_installed():
    """Test packages_installed()."""
    # list as input
    assert len(packages_installed([])) == 0
    assert len(packages_installed(['unknown-package'])) == 0
    assert len(packages_installed(['python3'])) == 1
    # tuples as input
    assert len(packages_installed(())) == 0
    assert len(packages_installed(('unknown-package', ))) == 0
    assert len(packages_installed(('python3', ))) == 1



# SPDX-License-Identifier: AGPL-3.0-or-later

from django.utils.translation import gettext_lazy as _

from plinth.clients import store_url

clients = [{
    'name': _('Transmission'),
    'platforms': [{
        'type': 'web',
        'url': '/transmission'
    }]
}, {
    'name':
        _('Tremotesf'),
    'platforms': [{
        'type': 'store',
        'os': 'android',
        'store_name': 'f-droid',
        'url': store_url('f-droid', 'org.equeim.tremotesf')
    }, {
        'type': 'store',
        'os': 'android',
        'store_name': 'google-play',
        'url': store_url('google-play', 'org.equeim.tremotesf')
    }]
}]

backup = {
    'data': {
        'directories': ['/var/lib/transmission-daemon/.config']
    },
    'secrets': {
        'files': ['/etc/transmission-daemon/settings.json']
    },
    'services': ['transmission-daemon']
}



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Django URLconf file containing all urls
"""
from captcha import views as cviews
from django.urls import include, re_path
from stronghold.decorators import public

from . import views

system_urlpatterns = [
    re_path(r'^sys/#visibility$', views.system_index, name='visibility'),
    re_path(r'^sys/#data$', views.system_index, name='data'),
    re_path(r'^sys/#system$', views.system_index, name='system'),
    re_path(r'^sys/#security$', views.system_index, name='security'),
    re_path(r'^sys/#administration$', views.system_index,
            name='administration'),
]

urlpatterns = [
    re_path(r'^$', views.index, name='index'),
    re_path(r'^language-selection/$',
            public(views.LanguageSelectionView.as_view()),
            name='language-selection'),
    re_path(r'^apps/$', views.AppsIndexView.as_view(), name='apps'),
    re_path(r'^sys/$', views.system_index, name='system'),
    re_path(r'', include((system_urlpatterns, 'system'))),
    re_path(r'^uninstall/(?P<app_id>[1-9a-z\-_]+)/$',
            views.UninstallView.as_view(), name='uninstall'),
    re_path(r'^rerun-setup/(?P<app_id>[1-9a-z\-_]+)/$', views.rerun_setup_view,
            name='rerun-setup'),

    # captcha urls are public
    re_path(r'^captcha/image/(?P<key>\w+)/$', public(cviews.captcha_image),
            name='captcha-image', kwargs={'scale': 1}),
    re_path(r'^captcha/image/(?P<key>\w+)@2/$', public(cviews.captcha_image),
            name='captcha-image-2x', kwargs={'scale': 2}),
    re_path(r'^captcha/audio/(?P<key>\w+)/$', public(cviews.captcha_audio),
            name='captcha-audio'),
    re_path(r'^captcha/refresh/$', public(cviews.captcha_refresh),
            name='captcha-refresh'),

    # Notifications
    re_path(r'^notification/(?P<id>[A-Za-z0-9-=]+)/dismiss/$',
            views.notification_dismiss, name='notification_dismiss')
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
URLs for the ikiwiki module
"""

from django.urls import re_path

from . import views

urlpatterns = [
    re_path(r'^apps/ikiwiki/$', views.IkiwikiAppView.as_view(), name='index'),
    re_path(r'^apps/ikiwiki/(?P<name>.+)/delete/$', views.delete,
            name='delete'),
    re_path(r'^apps/ikiwiki/create/$', views.create, name='create'),
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
URLs for Cockpit module.
"""

from django.urls import re_path

from plinth.views import AppView

urlpatterns = [
    re_path(r'^sys/cockpit/$', AppView.as_view(app_id='cockpit'),
            name='index'),
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
URLs for the email module.
"""

from django.urls import path
from stronghold.decorators import public

from plinth.utils import non_admin_view

from . import views

urlpatterns = [
    path('apps/email/', views.EmailAppView.as_view(), name='index'),
    path('apps/email/aliases/', non_admin_view(views.AliasView.as_view()),
         name='aliases'),
    path('apps/email/config.xml', public(views.XmlView.as_view())),
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Functional, browser based tests for RSS-Bridge app.
"""

import json
import subprocess

from plinth.tests import functional


class TestRSSBridgeApp(functional.BaseAppTests):
    app_name = 'rssbridge'
    has_service = False
    has_web = True

    def test_active_bridges(self, session_browser):
        """Check that bridges are active."""
        functional.app_enable(session_browser, self.app_name)
        functional.visit(session_browser, '/rss-bridge/')
        assert session_browser.find_by_css(
            '#bridge-Wikipedia button[type="submit"]')

    def test_feed_html(self, session_browser):
        """Check that a feed is properly rendered."""
        functional.app_enable(session_browser, self.app_name)
        url = '/rss-bridge/?action=display&bridge=Wikipedia&' \
            'language=en&subject=tfa&format=Html'
        functional.visit(session_browser, url)
        assert not session_browser.find_by_css('.exception-message')
        assert session_browser.find_by_css('.feeditem')

    def test_feed_json(self, session_browser):
        """Check that a feed is properly rendered."""
        functional.app_enable(session_browser, self.app_name)
        path = '/rss-bridge/?action=display&bridge=Wikipedia&' \
            'language=en&subject=tfa&format=Json'
        url = functional.config['DEFAULT']['url'] + path

        # Unauthorized
        result = subprocess.run(['curl', '-k', '--basic', url],
                                stdout=subprocess.PIPE,
                                stderr=subprocess.DEVNULL, check=True)
        assert '401 Unauthorized' in result.stdout.decode()

        # Authorized
        username = functional.config['DEFAULT']['username']
        password = functional.config['DEFAULT']['password']
        result = subprocess.run(
            ['curl', '-k', '--basic', '--user', f'{username}:{password}', url],
            stdout=subprocess.PIPE, stderr=subprocess.DEVNULL, check=True)
        feed = json.loads(result.stdout)
        assert len(feed['items'])



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
URLs for the minetest module.
"""

from django.urls import re_path

from plinth.modules.minetest.views import MinetestAppView

urlpatterns = [
    re_path(r'^apps/minetest/$', MinetestAppView.as_view(), name='index'),
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Functional, browser based tests for jsxc app."""

import pytest

from plinth.tests import functional

pytestmark = [pytest.mark.apps, pytest.mark.jsxc]


class TestJSXCApp(functional.BaseAppTests):
    app_name = 'jsxc'



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Functional, browser based tests for Kiwix app.
"""

import pathlib
from time import sleep

import pytest

from plinth.tests import functional

from .test_privileged import ZIM_ID

pytestmark = [pytest.mark.apps, pytest.mark.sso, pytest.mark.kiwix]

_default_url = functional.config['DEFAULT']['url']

_data_dir = pathlib.Path(__file__).parent / 'data'


class TestKiwixApp(functional.BaseAppTests):
    """Basic functional tests for Kiwix app."""

    app_name = 'kiwix'
    has_service = True
    has_web = True

    def test_add_delete_package(self, session_browser):
        """Test adding/deleting content package to the library."""
        functional.app_enable(session_browser, 'kiwix')

        zim_file = _data_dir / 'FreedomBox.zim'
        _add_package(session_browser, str(zim_file))
        assert _is_package_listed(session_browser, 'freedombox')
        assert _is_package_available(session_browser, 'FreedomBox')

        _delete_package(session_browser, ZIM_ID)
        assert not _is_package_listed(session_browser, 'freedombox')
        assert not _is_package_available(session_browser, 'FreedomBox')

    @pytest.mark.backups
    def test_backup_restore(self, session_browser):
        """Test backing up and restoring."""
        functional.app_enable(session_browser, 'kiwix')

        zim_file = _data_dir / 'FreedomBox.zim'
        _add_package(session_browser, str(zim_file))
        functional.backup_create(session_browser, 'kiwix', 'test_kiwix')

        _delete_package(session_browser, ZIM_ID)
        functional.backup_restore(session_browser, 'kiwix', 'test_kiwix')

        assert _is_package_listed(session_browser, 'freedombox')
        assert _is_package_available(session_browser, 'FreedomBox')

    def test_add_invalid_zim_file(self, session_browser):
        """Test handling of invalid zim files."""
        functional.app_enable(session_browser, 'kiwix')

        zim_file = _data_dir / 'invalid.zim'
        _add_package(session_browser, str(zim_file))

        assert not _is_package_listed(session_browser, 'invalid')

    def test_anonymous_access(self, session_browser):
        """Test anonymous access to Kiwix library."""
        functional.app_enable(session_browser, 'kiwix')
        functional.logout(session_browser)
        assert functional.eventually(_shortcut_exists, args=[session_browser])
        assert functional.eventually(_is_anonymous_read_allowed,
                                     args=[session_browser])


def _add_package(browser, file_name):
    """Add a package by uploading the ZIM file in kiwix app page."""
    browser.links.find_by_href('/plinth/apps/kiwix/package/add/').first.click()
    browser.attach_file('kiwix-file', file_name)
    functional.submit(browser, form_class='form-kiwix')


def _is_package_available(browser, title) -> bool:
    """Check whether a ZIM file is available in Kiwix web interface."""
    browser.visit(f'{_default_url}/kiwix')
    sleep(1)  # Allow time for the books to appear
    titles = browser.find_by_id('book__title')
    return any(element.value == title for element in titles)


def _is_package_listed(browser, name) -> bool:
    """Return whether a content package is list in kiwix app page."""
    functional.nav_to_module(browser, 'kiwix')
    links_found = browser.links.find_by_partial_href(f'/kiwix/viewer#{name}')
    return len(links_found) == 1


def _delete_package(browser, zim_id):
    """Delete a content package from the kiwix app page."""
    functional.nav_to_module(browser, 'kiwix')
    link = browser.links.find_by_href(
        f'/plinth/apps/kiwix/package/{zim_id}/delete/')
    if not link:
        raise ValueError('ZIM file missing!')

    link.first.click()
    functional.submit(browser, form_class='form-delete')


def _is_anonymous_read_allowed(browser) -> bool:
    """Check if Kiwix library page can be accessed without logging in."""
    functional.visit(browser, '/kiwix')
    return browser.is_element_present_by_id('kiwixfooter')


def _shortcut_exists(browser) -> bool:
    """Check that the Kiwix shortcut exists on the front page."""
    browser.visit(_default_url)
    links_found = browser.links.find_by_href('/kiwix')
    return len(links_found) == 1



# SPDX-License-Identifier: AGPL-3.0-or-later

from django.utils.translation import gettext_lazy as _

from plinth.clients import store_url

_package_id = 'com.nutomic.syncthingandroid'
_download_url = 'https://syncthing.net/'

clients = [{
    'name':
        _('Syncthing'),
    'platforms': [{
        'type': 'package',
        'format': 'deb',
        'name': 'syncthing',
    }, {
        'type': 'package',
        'format': 'brew',
        'name': 'syncthing',
    }, {
        'type': 'download',
        'os': 'gnu-linux',
        'url': _download_url,
    }, {
        'type': 'download',
        'os': 'macos',
        'url': _download_url,
    }, {
        'type': 'download',
        'os': 'windows',
        'url': _download_url,
    }, {
        'type': 'store',
        'os': 'android',
        'store_name': 'google-play',
        'url': store_url('google-play', _package_id)
    }, {
        'type': 'store',
        'os': 'android',
        'store_name': 'f-droid',
        'url': store_url('f-droid', _package_id)
    }, {
        'type': 'web',
        'url': '/syncthing'
    }]
}]

backup = {
    'secrets': {
        'directories': ['/var/lib/syncthing/.config']
    },
    'services': ['syncthing@syncthing']
}



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Forms for basic system configuration
"""

import logging
import re

from django import forms
from django.core import validators
from django.core.exceptions import ValidationError
from django.utils.translation import gettext as _
from django.utils.translation import gettext_lazy

from plinth import cfg, frontpage
from plinth.modules.apache import get_users_with_website
from plinth.utils import format_lazy

from . import home_page_url2scid

logger = logging.getLogger(__name__)

HOSTNAME_REGEX = r'^[a-zA-Z0-9]([-a-zA-Z0-9]{,61}[a-zA-Z0-9])?$'


def domain_label_validator(domainname):
    """Validate domain name labels."""
    for label in domainname.split('.'):
        if not re.match(HOSTNAME_REGEX, label):
            raise ValidationError(_('Invalid domain name'))


def get_homepage_choices():
    """Return list of drop down choices for home page."""
    shortcuts = frontpage.Shortcut.list(web_apps_only=True)
    shortcut_choices = [(shortcut.component_id, shortcut.name)
                        for shortcut in shortcuts if shortcut.is_enabled()]
    uws_choices = \
        [(home_page_url2scid(url),
         format_lazy(gettext_lazy("{user}'s website"), user=user))
         for user, url in get_users_with_website().items()]
    apache_default = ('apache-default', _('Apache Default'))
    plinth = ('plinth', _('FreedomBox Service (Plinth)'))
    return [apache_default, plinth] + uws_choices + shortcut_choices


class ConfigurationForm(forms.Form):
    """Main system configuration form"""
    # See:
    # https://tools.ietf.org/html/rfc952
    # https://tools.ietf.org/html/rfc1035#section-2.3.1
    # https://tools.ietf.org/html/rfc1123#section-2
    # https://tools.ietf.org/html/rfc2181#section-11
    hostname = forms.CharField(
        label=gettext_lazy('Hostname'), help_text=format_lazy(
            gettext_lazy(
                'Hostname is the local name by which other devices on the '
                'local network can reach your {box_name}.  It must start and '
                'end with an alphabet or a digit and have as interior '
                'characters only alphabets, digits and hyphens.  Total '
                'length must be 63 characters or less.'),
            box_name=gettext_lazy(cfg.box_name)), validators=[
                validators.RegexValidator(HOSTNAME_REGEX,
                                          gettext_lazy('Invalid hostname'))
            ], strip=True)

    domainname = forms.CharField(
        label=gettext_lazy('Domain Name'), help_text=format_lazy(
            gettext_lazy(
                'Domain name is the global name by which other devices on the '
                'Internet can reach your {box_name}.  It must consist of '
                'labels separated by dots.  Each label must start and end '
                'with an alphabet or a digit and have as interior characters '
                'only alphabets, digits and hyphens.  Length of each label '
                'must be 63 characters or less.  Total length of domain name '
                'must be 253 characters or less.'),
            box_name=gettext_lazy(cfg.box_name)), required=False, validators=[
                validators.RegexValidator(
                    r'^[a-zA-Z0-9]([-a-zA-Z0-9.]{,251}[a-zA-Z0-9])?$',
                    gettext_lazy('Invalid domain name')),
                domain_label_validator
            ], strip=True)

    homepage = forms.ChoiceField(
        label=gettext_lazy('Webserver Home Page'), help_text=format_lazy(
            gettext_lazy(
                'Choose the default page that must be served when '
                'someone visits your {box_name} on the web. A typical use '
                'case is to set your blog or wiki as the home page when '
                'someone visits the domain name. Note that once the home '
                'page is set to something other than {box_name} Service '
                '(Plinth), your users must explicitly type /plinth or '
                '/freedombox to reach {box_name} Service (Plinth).'),
            box_name=gettext_lazy(cfg.box_name)), required=False,
        choices=get_homepage_choices)

    advanced_mode = forms.BooleanField(
        label=gettext_lazy('Show advanced apps and features'), required=False,
        help_text=gettext_lazy(
            'Show apps and features that require more technical '
            'knowledge.'))

    logging_mode = forms.ChoiceField(
        label=gettext_lazy('System-wide logging'),
        choices=(('none', gettext_lazy('Disable logging, for privacy')),
                 ('volatile',
                  gettext_lazy('Keep some in memory until a restart, '
                               'for performance')),
                 ('persistent',
                  gettext_lazy('Write to disk, useful for debugging'))),
        help_text=gettext_lazy(
            'Logs contain information about who accessed the system and debug '
            'information from various services'))



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Common forms for use by modules.
"""

import os

from django import forms
from django.conf import settings
from django.utils import translation
from django.utils.translation import get_language_info
from django.utils.translation import gettext_lazy as _

import plinth


def _get_repository_choices():
    """Return the list of available repositories."""
    import plinth.modules.backups.repository as repository_module
    choices = [(repository.uuid, repository.name)
               for repository in repository_module.get_repositories()
               if repository.is_usable()]

    return choices


class AppEnableDisableForm(forms.Form):
    """Form to enable / disable an app."""
    should_enable = forms.BooleanField(widget=forms.HiddenInput,
                                       required=False)


class UninstallForm(forms.Form):
    """Form to uninstall an app."""
    should_backup = forms.BooleanField(
        label=_('Backup app before uninstall'),
        help_text=_('Restoring from the backup will restore app data.'),
        required=False, initial=True)
    repository = forms.ChoiceField(label=_('Repository to backup to'),
                                   choices=_get_repository_choices)

    def __init__(self, **kwargs):
        """Determine if backup fields must be used."""
        has_backup_restore = kwargs.pop('has_backup_restore', True)
        super().__init__(**kwargs)
        if not has_backup_restore:
            del self.fields['should_backup']
            del self.fields['repository']


class DomainSelectionForm(forms.Form):
    """Form for selecting a domain name to be used for
    distributed federated applications
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        from plinth.modules.names.components import DomainName
        domains = list(DomainName.list_names())
        self.fields['domain_name'].choices = zip(domains, domains)

    domain_name = forms.ChoiceField(
        label=_('Select a domain name to be used with this application'),
        help_text=_(
            'Warning! The application may not work properly if domain name is '
            'changed later.'), choices=[])


def _get_domain_choices():
    """Double domain entries for inclusion in the choice field."""
    from plinth.modules.names import get_available_tls_domains
    return ((domain, domain) for domain in get_available_tls_domains())


class TLSDomainForm(forms.Form):
    """Form to select a TLS domain for an app."""
    domain = forms.ChoiceField(
        choices=_get_domain_choices,
        label=_('TLS domain'),
        help_text=_(
            'Select a domain to use TLS with. If the list is empty, please '
            'configure at least one domain with certificates.'),
        required=False,
    )


class LanguageSelectionFormMixin:
    """Form mixin for selecting the user's preferred language."""

    language = forms.ChoiceField(
        label=_('Language'),
        help_text=_('Language to use for presenting this web interface'),
        required=False)

    def __init__(self, *args, **kwargs):
        """Initialize the form to fill language choice values."""
        super().__init__(*args, **kwargs)
        supported_languages = [
            (None, _('Use the language preference set in the browser'))
        ]

        def _get_local_name(language_code, language_name):
            try:
                return get_language_info(language_code)['name_local']
            except KeyError:
                return language_name

        for language_code, language_name in settings.LANGUAGES:
            locale_code = translation.to_locale(language_code)
            plinth_dir = os.path.dirname(plinth.__file__)
            if language_code == 'en' or os.path.exists(
                    os.path.join(plinth_dir, 'locale', locale_code)):
                supported_languages.append(
                    (language_code,
                     _get_local_name(language_code, language_name)))

        self.fields['language'].choices = supported_languages


class LanguageSelectionForm(LanguageSelectionFormMixin, forms.Form):
    """Language selection form."""

    language = LanguageSelectionFormMixin.language


class CheckboxSelectMultipleWithReadOnly(forms.widgets.CheckboxSelectMultiple):
    """Multiple checkbox widget that allows setting individual fields readonly.

    To mark a feature as readonly an option, pass a dict instead of a string
    for its label, of the form: {'label': 'option label', 'disabled': True}.

    """

    def create_option(self, name, value, label, selected, index, subindex=None,
                      attrs=None):
        option = super().create_option(name, value, label, selected, index,
                                       subindex, attrs)
        if isinstance(option['label'], dict):
            if option['label'].get('disabled'):
                option['attrs']['disabled'] = 'disabled'

            option['label'] = option['label']['label']

        return option


class CheckboxSelectMultiple(forms.widgets.CheckboxSelectMultiple):
    """CheckboxSelectMultiple with ids named after choices."""

    def id_for_label(self, id_, index=None):
        """
        Make ids looks like id_groups_groupname where
        a choice is like ('groupname', 'Group description').
        """
        if index is None:
            return ''
        if id_ and self.add_id_index:
            id_ = '%s_%s' % (id_, list(self.choices)[int(index)][0])
        return id_



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app to configure Single Sign On services."""

from django.utils.translation import gettext_lazy as _

from plinth import app as app_module
from plinth.config import DropinConfigs
from plinth.package import Packages

from . import privileged


class SSOApp(app_module.App):
    """FreedomBox app for single sign on."""

    app_id = 'sso'

    _version = 2

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()

        info = app_module.Info(app_id=self.app_id, version=self._version,
                               is_essential=True,
                               depends=['security',
                                        'apache'], name=_('Single Sign On'))
        self.add(info)

        packages = Packages('packages-sso', [
            'libapache2-mod-auth-pubtkt', 'openssl', 'python3-openssl', 'flite'
        ])
        self.add(packages)

        dropin_configs = DropinConfigs('dropin-configs-sso', [
            '/etc/apache2/includes/freedombox-single-sign-on.conf',
        ])
        self.add(dropin_configs)

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)
        privileged.create_key_pair()



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app for configuring Zoph photo organiser."""

import logging

from django.contrib import messages
from django.http import HttpResponseRedirect
from django.shortcuts import redirect
from django.urls import reverse_lazy
from django.utils.translation import gettext as _
from django.views.generic import TemplateView

from plinth import app as app_module
from plinth import views

from . import privileged
from .forms import ZophForm

logger = logging.getLogger(__name__)


class SetupView(TemplateView):
    """Show zoph setup page."""

    template_name = 'zoph-pre-setup.html'
    success_url = reverse_lazy('zoph:index')

    def get_context_data(self, *args, **kwargs):
        """Provide context data to the template."""
        context = super().get_context_data(**kwargs)
        app = app_module.App.get('zoph')
        context['title'] = app.info.name
        context['app_info'] = app.info
        return context

    def post(self, _request, *args, **kwargs):
        """Handle form submission."""
        admin_user = self.request.user.get_username()
        privileged.set_configuration(admin_user=admin_user)
        return HttpResponseRedirect(reverse_lazy('zoph:index'))


class ZophAppView(views.AppView):
    """App configuration page."""

    form_class = ZophForm
    app_id = 'zoph'

    def dispatch(self, request, *args, **kwargs):
        """Redirect to setup page if setup is not done yet."""
        status = self.get_common_status()
        if status['is_enabled']:
            # When disabled, such as when uninstalling
            if privileged.is_configured() is False:
                return redirect('zoph:setup')

        return super().dispatch(request, *args, **kwargs)

    def get_initial(self):
        """Get the current settings from Zoph."""
        status = super().get_initial()
        config = privileged.get_configuration()
        status['enable_osm'] = (config['maps.provider'] == 'osm')
        return status

    def form_valid(self, form):
        """Apply the changes submitted in the form."""
        old_status = form.initial
        new_status = form.cleaned_data
        if old_status['enable_osm'] != new_status['enable_osm']:
            try:
                privileged.set_configuration(
                    enable_osm=new_status['enable_osm'])
                messages.success(self.request, _('Configuration updated.'))
            except Exception:
                messages.error(self.request,
                               _('An error occurred during configuration.'))

        return super().form_valid(form)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Discover, load and manage FreedomBox applications.
"""

import collections
import importlib
import logging
import pathlib
import re

import django

from plinth import cfg
from plinth.signals import pre_module_loading

logger = logging.getLogger(__name__)

loaded_modules = collections.OrderedDict()
_modules_to_load = None


def include_urls():
    """Include the URLs of the modules into main Django project."""
    for module_import_path in get_modules_to_load():
        module_name = module_import_path.split('.')[-1]
        _include_module_urls(module_import_path, module_name)


def load_modules():
    """
    Read names of enabled modules in modules/enabled directory and
    import them from modules directory.
    """
    pre_module_loading.send_robust(sender="module_loader")
    for module_import_path in get_modules_to_load():
        module_name = module_import_path.split('.')[-1]
        try:
            loaded_modules[module_name] = importlib.import_module(
                module_import_path)
        except Exception as exception:
            logger.exception('Could not import %s: %s', module_import_path,
                             exception)
            if cfg.develop:
                raise


def _include_module_urls(module_import_path, module_name):
    """Include the module's URLs in global project URLs list"""
    from plinth import urls
    url_module = module_import_path + '.urls'
    try:
        urls.urlpatterns += [
            django.urls.re_path(r'',
                                django.urls.include((url_module, module_name)))
        ]
    except ImportError:
        logger.debug('No URLs for %s', module_name)
        if cfg.develop:
            raise


def _get_modules_enabled_paths():
    """Return list of paths from which enabled modules list must be read."""
    return [
        pathlib.Path('/usr/share/freedombox/modules-enabled/'),
        pathlib.Path('/etc/plinth/modules-enabled/'),  # For compatibility
        pathlib.Path('/etc/freedombox/modules-enabled/'),
    ]


def _get_modules_enabled_files_to_read():
    """Return the list of files containing modules import paths."""
    module_files = {}
    for path in _get_modules_enabled_paths():
        directory = pathlib.Path(path)
        files = list(directory.glob('*'))
        for file_ in files:
            # Omit hidden or backup files
            if file_.name.startswith('.') or '.dpkg' in file_.name:
                continue

            if file_.is_symlink() and str(file_.resolve()) == '/dev/null':
                module_files.pop(file_.name, None)
                continue

            module_files[file_.name] = file_

    if module_files:
        return module_files.values()

    # 'make build install' has not been executed yet. Pickup files to load from
    # local module directories.
    directory = pathlib.Path(__file__).parent
    glob_pattern = 'modules/*/data/usr/share/freedombox/modules-enabled/*'
    return list(directory.glob(glob_pattern))


def get_modules_to_load():
    """Get the list of modules to be loaded"""
    global _modules_to_load
    if _modules_to_load is not None:
        return _modules_to_load

    modules = []
    for file_ in _get_modules_enabled_files_to_read():
        module = _read_module_import_paths_from_file(file_)
        if module:
            modules.append(module)

    _modules_to_load = modules
    return modules


def get_module_import_path(module_name: str) -> str:
    """Return the import path for a module."""
    import_path_file = None
    for path in _get_modules_enabled_paths():
        file_ = path / module_name
        if file_.exists():
            import_path_file = file_

        if file_.is_symlink() and str(file_.resolve()) == '/dev/null':
            import_path_file = None

    if not import_path_file:
        # 'make build install' has not been executed yet. Pickup files to load
        # from local module directories.
        directory = pathlib.Path(__file__).parent
        import_path_file = (directory /
                            f'modules/{module_name}/data/usr/share/'
                            f'freedombox/modules-enabled/{module_name}')

    if not import_path_file:
        raise ValueError('Unknown module')

    import_path = _read_module_import_paths_from_file(import_path_file)
    if not import_path:
        raise ValueError('Module disabled')

    return import_path


def _read_module_import_paths_from_file(file_path: pathlib.Path) -> str | None:
    """Read a module's import path from a file."""
    with file_path.open() as file_handle:
        for line in file_handle:
            line = re.sub('#.*', '', line)
            line = line.strip()
            if line:
                return line

    return None



# SPDX-License-Identifier: AGPL-3.0-or-later

from django.utils.translation import gettext_lazy as _

from plinth.clients import store_url

clients = [
    {
        'name':
            _('Tiny Tiny RSS (TTTRSS)'),
        'platforms': [{
            'type': 'store',
            'os': 'android',
            'store_name': 'f-droid',
            'url': store_url('f-droid', 'org.fox.tttrss')
        }]
    },
    {
        'name':
            _('TTRSS-Reader'),
        'platforms': [{
            'type': 'store',
            'os': 'android',
            'store_name': 'google-play',
            'url': store_url('google-play', 'org.ttrssreader')
        }, {
            'type': 'store',
            'os': 'android',
            'store_name': 'f-droid',
            'url': store_url('f-droid', 'org.ttrssreader')
        }]
    },
    {
        'name':
            _('Geekttrss'),
        'platforms': [{
            'type': 'store',
            'os': 'android',
            'store_name': 'google-play',
            'url': store_url('google-play', 'com.geekorum.ttrss')
        }]
    },
    {
        'name': _('Tiny Tiny RSS'),
        'platforms': [{
            'type': 'web',
            'url': '/tt-rss'
        }]
    },
]

backup = {
    'data': {
        'files': ['/var/lib/plinth/backups-data/ttrss-database.sql']
    },
    'secrets': {
        'files': [
            '/etc/tt-rss/database.php',
            '/etc/dbconfig-common/tt-rss.conf',
        ]
    },
    'services': ['tt-rss']
}



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
URLs for the Searx module.
"""

from django.urls import re_path

from .views import SearxAppView

urlpatterns = [
    re_path(r'^apps/searx/$', SearxAppView.as_view(), name='index'),
]
